from tinydb import TinyDB, Query
from typing import List, Dict
from threading import RLock
from datetime import datetime
from .singleton import SingletonMeta
from .tinydb_exceptions import *
import vrmjobs


class TinyDbWrapper(metaclass=SingletonMeta):
    """
    Wrapper class for TinyDB within VRM system
    """
    def __init__(self, db_path: str) -> None:
        # create db
        self.db = TinyDB(db_path, sort_keys=True, indent=2, separators=(',', ': '))
        # create worker table
        self.table = self.db.table('hosts')
        # create lock
        self.lock = RLock()
        # create timing-format
        self.time_format = '%H:%M:%S %d-%m-%Y'

    def insert_host(self, info: 'vrmjobs.HostInfo') -> bool:
        """
        Insert a new host into db
        :param info: an instance of HostInfo
        :return: True if success, False otherwise
        """
        with self.lock:
            try:
                host = self._check_host_existence(info.hostname)

                if not host:
                    ports = []
                    for p in info.ports:
                        ports.append({"daemon": p.daemon, "port": p.port})

                    self.table.insert({"hostname": info.hostname,
                                       "inet_addr": info.inet_addr,
                                       "ports": ports,
                                       "type": info.type.name,
                                       "latest_recv": datetime.now().strftime(self.time_format)})
                    return True
                return False
            except Exception as err:
                raise InsertError('Cannot insert new host {}'.format(str(info)), err)

    def _check_host_existence(self, hostname: str) -> bool:
        """
        Check whether a host with certain hostname is already included
        :param hostname: hostname of the host need to check
        :return: True if included, False otherwise
        """
        with self.lock:
            hosts = self.table.all()
            for host in hosts:
                if host['hostname'] == hostname:
                    return True
            return False

    def get_host_by_hostname(self, hostname: str) -> vrmjobs.HostInfo:
        """
        Get info of a host by its hostname
        :param hostname: hostname of a worker/controller/monitor
        :return: HostInfo instance
        """
        with self.lock:
            try:
                host = Query()
                record = self.table.search(host.hostname == hostname)

                if record:
                    port_infos = []
                    for info in record[0]["ports"]:
                        port_infos.append(vrmjobs.PortInfo(info["daemon"], info["port"]))

                    return vrmjobs.HostInfo(record[0]["hostname"],
                                            record[0]["inet_addr"],
                                            port_infos,
                                            # https://stackoverflow.com/questions/41407414/convert-string-to-enum-in-python
                                            vrmjobs.HostType.__dict__[record[0]["type"]])
                else:
                    raise Exception('Not found.')
            except Exception as err:
                raise GetError('Cannot find host by hostname={}'.format(hostname), err)

    def get_daemon_by_name(self, hostname: str, daemon: str) -> vrmjobs.PortInfo:
        """
        Get daemon information of a worker/collector/monitor by its hostname
        :param hostname: hostname of a host
        :param daemon: name of daemon
        :return: an instance of PortInfo
        """
        with self.lock:
            try:
                host = Query()
                record = self.table.search(host.hostname == hostname)[0]

                if not record:
                    return None

                for info in record["ports"]:
                    if info["daemon"] == daemon:
                        return vrmjobs.PortInfo(info["daemon"],
                                                info["port"])

                return None
            except Exception as err:
                raise GetError('Cannot get daemon name={} of host with hostname={}'.format(daemon, hostname), err)

    def update_host_metrics(self, hostname, metrics: List[Dict]) -> bool:
        """
        Update metrics information of a host with a certain hostname
        :param hostname: hostname of host
        :param metrics: list of PortInfo
        :return: True if succeed, False otherwise
        """
        with self.lock:
            try:
                host = Query()
                self.table.upsert({'metrics': metrics},
                                  host.hostname.matches(hostname))

                return True
            except Exception as err:
                raise UpdateError('Cannot update metrics: {} of host with hostname={}'.format(metrics, hostname), err)

    def update_host_heartbeat(self, hostname: str) -> bool:
        """
        Update latest_update information of a host with a certain hostname
        :param hostname: hostname of host
        :return: True if success, UpdateError otherwise
        """
        with self.lock:
            try:
                host = Query()
                self.table.update({'latest_recv': datetime.now().strftime(self.time_format)},
                                  host.hostname.matches(hostname))
                return True
            except Exception as err:
                raise UpdateError('Cannot update latest_recv of host with hostname={}'.format(hostname), err)

    def check_heartbeat(self, hostname: str, current_time, max_interval: int) -> bool:
        """
        Check whether latest_update information of a host with a certain hostname is overdue
        :param hostname: hostname of host
        :param current_time: time of the moment of checking
        :param max_interval: maximum seconds allowed between check
        :return:
        """
        with self.lock:
            try:
                host = Query()
                record = self.table.search(host.hostname == hostname)[0]

                if record:
                    latest_update = datetime.strptime(record['latest_recv'], self.time_format)
                    difference = current_time - latest_update

                    return difference.total_seconds() <= max_interval

            except Exception as err:
                raise HeartbeatError('Cannot check heartbeat of host with hostname={}'.format(hostname), err)



# singleton class
from .singleton import SingletonMeta
# tinydb wrapper
from .tinydb_wrapper import TinyDbWrapper
from .tinydb_exceptions import *
# static helper methods/functions
from .helper import *


from utilities import *
import vrmjobs
from datetime import datetime


def test_check_heartbeat(db_manager: 'TinyDbWrapper', hostname: str, interval: int):
    try:
        result = db_manager.check_heartbeat(hostname, datetime.now(), interval)

        if result:
            print("Still good.")
        else:
            print("Need to update.")
    except HeartbeatError as err:
        print(err)


def test_update_heartbeat(db_manager: 'TinyDbWrapper', hostname: str):
    try:
        db_manager.update_host_heartbeat(hostname)
        host = db_manager.get_host_by_hostname(hostname)
        print("Latest info: {}".format(host))
    except UpdateError as err:
        print(err)


def test_get_host(db_manager: 'TinyDbWrapper', hostname: str):
    try:
        host = db_manager.get_host_by_hostname(hostname)
        print(host)
    except GetError as err:
        print(err)


def test_insert_hosts(db_manager: 'TinyDbWrapper', hostname: str, inet_addr:str,
                      ports: ['vrmjobs.PortInfo'], hosttype: 'vrmjobs.HostType'):
    try:
        host = vrmjobs.HostInfo(hostname, inet_addr, ports, hosttype)
        result = db_manager.insert_host(host)
        if result:
            print("Insert {} to db successfully.".format(host))

    except InsertError as err:
        print(err)


def main():
    # create db
    db = TinyDbWrapper('test_db.json')

    # insert a new host (if possible)
    #test_insert_hosts(db, 'worker1', '192.168.1.11', [vrmjobs.PortInfo('workerd', 10300),
    #                            vrmjobs.PortInfo('builderd', 5000)], vrmjobs.HostType.WORKER)

    #test_insert_hosts(db, 'worker2', '192.168.1.12', [vrmjobs.PortInfo('workerd', 10300),
    #                                                  vrmjobs.PortInfo('builderd', 5000)], vrmjobs.HostType.WORKER)

    #test_get_host(db, 'worker1')
    #test_get_host(db, 'worker3')
    #test_update_heartbeat(db, 'worker1')
    #test_update_heartbeat(db, 'worker2')
    test_check_heartbeat(db, 'worker1', 3600)


if __name__ == '__main__':
    main()



import re


def volume_convert_yaml_to_json(yaml_conf, version: str):
    """
    Convert yaml configuration to json used to create docker volume.
    Return array of dict, each dict is used in dcli.create_volume(...)
    :param yaml_conf: yaml configuration
    :param version: yaml version (support 2.x/3.x)
    """
    volume_conf_list = []
    # handle version 2
    #if re.match('^2\.(\d*)$', version):
    #    print("Convert volume yaml of version 2.x")
    # handle version 3
    if re.match('^3\.(\d*)$', version):
        for tup in yaml_conf.items():
            volume_conf = {'name': tup[0], 'driver_opts': {}, 'labels': {}}
            if not tup[1]:
                volume_conf['driver'] = 'local'

            if type(tup[1]) is dict:
                # check external
                try:
                    volume_conf['external'] = tup[1]['external']
                except KeyError:
                    pass

                # check driver
                try:
                    volume_conf['driver'] = tup[1]['driver']
                except KeyError:
                    volume_conf['driver'] = 'local'

                # check driver_opts
                try:
                    for dr_tup in tup[1]['driver_opts'].items():
                        volume_conf['driver_opts'][dr_tup[0]] = dr_tup[1]
                except KeyError:
                    pass

                # check labels
                try:
                    for l_tup in tup[1]['labels'].items():
                        volume_conf['labels'][l_tup[0]] = l_tup[1]
                except KeyError:
                    pass

            volume_conf_list.append(volume_conf)
        return volume_conf_list

    return {}


def network_convert_yaml_to_json(yaml_conf, version: str):
    """
    Convert yaml configuration to json used to create docker network.
    Return array of dict, each dict is used in dcli.create_network(...)
    :param yaml_conf: yaml configuration
    :param version: yaml version (support 2.x/3.x)
    """
    network_conf_list = []

    # handle version 2
    #if re.match('^2\.(\d*)$', version):
    #    print("Convert network yaml of version 2.x")

    # handle version 3
    if re.match('^3\.(\d*)$', version):
        for tup in yaml_conf.items():
            network_conf = {'name': tup[0], 'driver': None, 'options': None,
                            'ipam': None, 'check_duplicate': None, 'internal': False,
                            'labels': None, 'enable_ipv6': False, 'attachable': None,
                            'scope': None, 'ingress': None}

            if type(tup[1]) is dict:
                try:
                    network_conf['external'] = tup[1]['external']
                    network_conf_list.append(network_conf)
                    continue
                except KeyError:
                    pass

                # get driver of network
                try:
                    network_conf['driver'] = tup[1]['driver']
                except KeyError:
                    pass

                # get opts of driver
                try:
                    network_conf['options'] = {}
                    for dr_tup in tup[1]['driver_opts'].items():
                        network_conf['options'][dr_tup[0]] = dr_tup[1]
                except KeyError:
                    network_conf['options'] = None
                    pass

                # get ipam config
                try:
                    if tup[1]['ipam']:
                        ipam_conf = {'driver': 'default', 'pool_configs': [], 'options': None}
                        for ipam_tup in tup[1]['ipam'].items():
                            if ipam_tup[0] == 'driver':
                                ipam_conf['driver'] = ipam_tup[1]
                            if ipam_tup[0] == 'config':
                                for pool in ipam_tup[1]:
                                    pool_conf = {}
                                    if 'subnet' in pool:
                                        pool_conf['subnet'] = pool['subnet']
                                    if 'iprange' in pool:
                                        pool_conf['iprange'] = pool['iprange']
                                    if 'gateway' in pool:
                                        pool_conf['gateway'] = pool['gateway']
                                    if 'aux_addresses' in pool:
                                        pool_conf['aux_addresses'] = pool['aux_addresses']
                                    ipam_conf['pool_configs'].append(pool_conf)

                            if ipam_tup[0] == 'options':
                                ipam_conf['options'] = ipam_tup[1]
                        network_conf['ipam'] = ipam_conf

                except KeyError:
                    pass

                # get attachable
                try:
                    network_conf['attachable'] = tup[1]['attachable']
                except KeyError:
                    pass

                # get internal
                try:
                    network_conf['internal'] = tup[1]['internal']
                except KeyError:
                    pass

            network_conf_list.append(network_conf)
        return network_conf_list

    return {}


def container_convert_yaml_to_json(yaml_conf, version: str):
    """
    Convert yaml configuration to json used to create docker network.
    Return array of dict, each dict is used in dcli.create_container(...)
    :param yaml_conf: yaml configuration
    :param version: yaml version (support 2.x/3.x)
    """
    container_conf_list = []

    # handle version 2
    if re.match('^2\.(\d*)$', version):
        print("Convert container yaml of version 2.x")

    # handle version 3
    if re.match('^3\.(\d*)$', version):
        print("Convert container yaml of version 3.x")
        for tup in yaml_conf.items():
            container_conf = {
                'container_name': None,
                'image': None,
                'volumes': [],
                'command': [],
                'entrypoint': [],
                'network_mode': None,
                'networks': [],
                'labels': {},
                'expose': [],
                'extra_hosts': [],
                'dns': [],
                'dns_search': [],
                'environment': {},
                'ports': [],
                'restart': None,
                'sysctls': {},
                'tmpfs': [],
                'pid': None,
                'hostname': None,
                'privileged': False,
                'read_only': False
            }

            if type(tup[1]) is dict:
                # get container_name:
                try:
                    container_conf['container_name'] = tup[1]['container_name']
                except KeyError:
                    pass

                # get docker image
                try:
                    container_conf['image'] = tup[1]['image']
                except KeyError:
                    pass

                # get volumes:
                try:
                    for mount_cfg in tup[1]['volumes']:
                        container_conf['volumes'].append(mount_cfg)
                except KeyError:
                    pass

                # get docker commands
                try:
                    for cmd in tup[1]['command']:
                        container_conf['command'].append(cmd)
                except KeyError:
                    pass

                # get entrypoint
                try:
                    container_conf['entrypoint'] = tup[1]['entrypoint']
                except KeyError:
                    pass

                # get network_mode
                try:
                    container_conf['network_mode'] = tup[1]['network_mode']
                except KeyError:
                    pass

                # get networks
                try:
                    for network in tup[1]['networks']:
                        for net_tup in network.items():
                            net_conf = {'name': net_tup[0], 'config': {
                                'ipv4_address': None,
                                'ipv6_address': None,
                                'aliases': None
                            }}

                            if type(net_tup[1]) is dict:
                                try:
                                    net_conf['config']['ipv4_address'] = net_tup[1]['ipv4_address']
                                except KeyError:
                                    pass

                                try:
                                    net_conf['config']['ipv6_address'] = net_tup[1]['ipv6_address']
                                except KeyError:
                                    pass

                                try:
                                    net_conf['config']['aliases'] = net_tup[1]['aliases']
                                except KeyError:
                                    pass

                            container_conf['networks'].append(net_conf)
                except KeyError:
                    pass

                # get labels
                try:
                    for l_tup in tup[1]['labels'].items():
                        container_conf['labels'][l_tup[0]] = l_tup[1]
                except KeyError:
                    pass

                # get exposed ports
                try:
                    for port in tup[1]['expose']:
                        container_conf['expose'].append(int(port))
                except KeyError:
                    pass

                # get extra_hosts
                try:
                    for host in tup[1]['extra_hosts']:
                        container_conf['extra_hosts'].append(host)
                except KeyError:
                    pass

                # get dns
                try:
                    for ip in tup[1]['dns']:
                        container_conf['dns'].append(ip)
                except KeyError:
                    pass

                # get dns_search
                try:
                    for search in tup[1]['dns_search']:
                        container_conf['dns_search'].append(search)
                except KeyError:
                    pass

                # get environment
                try:
                    for env_tup in tup[1]['environment']:
                        container_conf['environment'][env_tup[0]] = env_tup[1]
                except KeyError:
                    pass

                # get ports
                try:
                    for port in tup[1]['ports']:
                        port_tup = tuple(part for part in port.split(':') if part)

                        # port bindings with specified host ip address
                        if len(port_tup) == 3:
                            # port bindings with protocol
                            if '/' in port_tup[2]:
                                proto_tup = tuple(part for part in port_tup[2].split(':') if part)
                                container_conf['ports'].append({
                                    'host_ip': port_tup[0],
                                    'host_port': int(port_tup[1]),
                                    'cont_port': int(proto_tup[0]),
                                    'proto': proto_tup[1]
                                })
                            else:
                                container_conf['ports'].append({
                                    'host_ip': port_tup[0],
                                    'host_port': int(port_tup[1]),
                                    'cont_port': int(port_tup[0])
                                })
                        # port bindings without specified host ip address
                        if len(port_tup) == 2:
                            # port bindings with protocol
                            if '/' in port_tup[1]:
                                proto_tup = tuple(part for part in port_tup[2].split(':') if part)
                                container_conf['ports'].append({
                                    'host_port': int(port_tup[0]),
                                    'cont_port': int(proto_tup[0]),
                                    'proto': proto_tup[1]
                                })
                            else:
                                container_conf['ports'].append({
                                    'host_port': int(port_tup[0]),
                                    'cont_port': int(port_tup[0])
                                })

                        # port exposed
                        if len(port_tup) == 1:
                            container_conf['ports'].append({
                                'host_port': None,
                                'cont_port': port_tup[0]
                            })

                except KeyError:
                    pass

                # get restart_policy
                try:
                    container_conf['restart'] = tup[1]['restart']
                except KeyError:
                    pass

                # get sysctls
                try:
                    for sys_tup in tup[1]['sysctls'].items():
                        container_conf['sysctls'][sys_tup[0]] = sys_tup[1]
                except KeyError:
                    pass

                # get tmpfs
                try:
                    container_conf['tmpfs'] = tup[1]['tmpfs']
                except KeyError:
                    pass

                # get pid
                try:
                    container_conf['pid'] = tup[1]['pid']
                except KeyError:
                    pass

                # get hostname
                try:
                    container_conf['hostname'] = tup[1]['hostname']
                except KeyError:
                    pass

                # get privileged
                try:
                    container_conf['privileged'] = tup[1]['privileged']
                except KeyError:
                    pass

                # get read_only
                try:
                    container_conf['read_only'] = tup[1]['read_only']
                except KeyError:
                    pass

                container_conf_list.append(container_conf)

    return container_conf_list



class InsertError(Exception):
    def __init__(self, msg, original_exception):
        super(InsertError, self).__init__("{} : {}".format(msg, original_exception))
        self.original_exception = original_exception


class GetError(Exception):
    def __init__(self, msg, original_exception):
        super(GetError, self).__init__("{} : {}".format(msg, original_exception))
        self.original_exception = original_exception


class RemoveError(Exception):
    def __init__(self, msg, original_exception):
        super(RemoveError, self).__init__("{} : {}".format(msg, original_exception))
        self.original_exception = original_exception


class UpdateError(Exception):
    def __init__(self, msg, original_exception):
        super(UpdateError, self).__init__("{} : {}".format(msg, original_exception))
        self.original_exception = original_exception


class HeartbeatError(Exception):
    def __init__(self, msg, original_exception):
        super(HeartbeatError, self).__init__("{} : {}".format(msg, original_exception))
        self.original_exception = original_exception



from threading import Lock


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]
